package net.knarcraft.ffmpegconverter.converter;

import net.knarcraft.ffmpegconverter.streams.AudioStream;
import net.knarcraft.ffmpegconverter.streams.StreamObject;
import net.knarcraft.ffmpegconverter.streams.SubtitleStream;
import net.knarcraft.ffmpegconverter.streams.VideoStream;
import net.knarcraft.ffmpegconverter.utility.FFMpegHelper;
import net.knarcraft.ffmpegconverter.utility.FileUtil;
import net.knarcraft.ffmpegconverter.utility.OutputUtil;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

/**
 * Implements all methods which can be useful for any implementation of a converter.
 */
public abstract class AbstractConverter implements Converter {
    final boolean debug = false;
    private final String outputExtension;
    String ffprobePath;
    String ffmpegPath;
    String[] audioFormats;
    String[] videoFormats;

    /**
     * Initializes variables used by the abstract converter
     */
    AbstractConverter(String outputExtension) {
        this.outputExtension = outputExtension;
        OutputUtil.setDebug(this.debug);
        try {
            audioFormats = FileUtil.readFileLines("audio_formats.txt");
            videoFormats = FileUtil.readFileLines("video_formats.txt");
        } catch (IOException e) {
            e.printStackTrace();
            System.exit(1);
        }
    }

    /**
     * Filters parsed streams into one of the stream types
     *
     * @param streams <p>A list of stream objects.</p>
     * @param clazz   <p>The class to filter</p>
     * @param <G>     <p>The correct object type for the streams with the selected codec type.</p>
     * @return <p>A potentially shorter list of streams.</p>
     */
    @SuppressWarnings("unchecked")
    static <G extends StreamObject> List<G> filterStreamsByType(List<StreamObject> streams, Class<?> clazz) {
        List<G> newStreams = new ArrayList<>();
        for (StreamObject stream : streams) {
            if (stream.getClass() == clazz) {
                newStreams.add((G) stream);
            }
        }
        return newStreams;
    }

    /**
     * Filters and sorts audio streams according to chosen languages
     *
     * @param audioStreams   <p>A list of audio streams.</p>
     * @param audioLanguages <p>A list of languages.</p>
     * @return <p>A list containing just audio tracks of chosen languages, sorted in order of languages.</p>
     */
    static List<AudioStream> filterAudioStreams(List<AudioStream> audioStreams, String[] audioLanguages) {
        return sortStreamsByLanguage(audioStreams, audioLanguages);
    }

    /**
     * Filters and sorts subtitle streams according to chosen languages
     *
     * @param subtitleStreams      <p>A list of subtitle streams.</p>
     * @param subtitleLanguages    <p>A list of languages.</p>
     * @param preventSignsAndSongs <p>Whether partial subtitles should be avoided.</p>
     * @return <p>A list containing just subtitles of chosen languages, sorted in order of languages.</p>
     */
    static List<SubtitleStream> filterSubtitleStreams(List<SubtitleStream> subtitleStreams, String[] subtitleLanguages,
                                                      boolean preventSignsAndSongs) {
        List<SubtitleStream> sorted = sortStreamsByLanguage(subtitleStreams, subtitleLanguages);
        sorted.removeIf((stream) -> preventSignsAndSongs && !stream.getIsFullSubtitle());
        return sorted;
    }

    /**
     * Sorts subtitle streams according to chosen languages and removes non matching languages
     *
     * @param streams   <p>A list of streams to sort.</p>
     * @param languages <p>The languages chosen by the user.</p>
     * @param <G>       <p>The type of streams to sort.</p>
     * @return <p>A sorted version of the list.</p>
     */
    private static <G extends StreamObject> List<G> sortStreamsByLanguage(List<G> streams, String[] languages) {
        List<G> sorted = new ArrayList<>();
        for (String language : languages) {
            for (G stream : streams) {
                String streamLanguage = stream.getLanguage();
                if (language.equals("*") || ((streamLanguage == null || streamLanguage.equals("und")) &&
                        language.equals("0")) || (streamLanguage != null && streamLanguage.equals(language))) {
                    sorted.add(stream);
                }
            }
            streams.removeAll(sorted);
        }
        return sorted;
    }

    /**
     * Gets all valid input formats for the converter
     *
     * @return <p>A list of valid input formats</p>
     */
    public abstract String[] getValidFormats();

    /**
     * Reads streams from a file, and converts it to an mp4
     *
     * @param folder <p>The folder of the file to process.</p>
     * @param file   <p>The file to process.</p>
     * @throws IOException <p>If the BufferedReader fails.</p>
     */
    private void processFile(File folder, File file) throws IOException {
        List<StreamObject> streams = FFMpegHelper.probeFile(ffprobePath, file);
        if (streams.isEmpty()) {
            throw new IllegalArgumentException("The file has no valid streams. Please make sure the file exists and" +
                    " is not corrupt.");
        }
        String newPath = FileUtil.getNonCollidingPath(folder, file, outputExtension);
        OutputUtil.println();
        OutputUtil.println("Preparing to start process...");
        OutputUtil.println("Converting " + file);
        ProcessBuilder processBuilder = new ProcessBuilder(generateConversionCommand(ffmpegPath, file, streams, newPath));
        FFMpegHelper.runProcess(processBuilder, folder, "\n", true);
    }

    /**
     * Gets the first audio stream from a list of streams
     *
     * @param audioStreams <p>A list of all streams.</p>
     * @return <p>The first audio stream found or null if no audio streams were found.</p>
     */
    AudioStream getFirstAudioStream(List<AudioStream> audioStreams) {
        AudioStream audioStream = null;
        if (audioStreams.size() > 0) {
            audioStream = audioStreams.get(0);
        }
        return audioStream;
    }

    /**
     * Gets the first subtitle stream from a list of streams
     *
     * @param subtitleStreams <p>A list of all subtitle streams.</p>
     * @return <p>The first subtitle stream found or null if no subtitle streams were found.</p>
     */
    SubtitleStream getFirstSubtitleStream(List<SubtitleStream> subtitleStreams) {
        SubtitleStream subtitleStream = null;
        if (subtitleStreams.size() > 0) {
            subtitleStream = subtitleStreams.get(0);
        }
        return subtitleStream;
    }

    /**
     * Gets the first video stream from a list of streams
     *
     * @param videoStreams <p>A list of all streams.</p>
     * @return <p>The first video stream found or null if no video streams were found.</p>
     */
    VideoStream getFirstVideoStream(List<VideoStream> videoStreams) {
        VideoStream videoStream = null;
        if (videoStreams.size() > 0) {
            videoStream = videoStreams.get(0);
        }
        if (videoStream == null) {
            throw new IllegalArgumentException("The file does not have any valid video streams.");
        }
        return videoStream;
    }

    @Override
    public void convert(File file) throws IOException {
        processFile(file.getParentFile(), file);
    }
}
